#include <glob.h>
#include <vector>
#include <string>
#include <fstream>
#include <cstdint>
#include <cmath>
#include <x86intrin.h>
#include "Utilities.h"

using namespace cv;
using namespace std;
using namespace Eigen;

namespace Utilities {
	void bilinearInterp_SIMD(Mat& Source, int X, int Y, double centerX, double centerY, Mat& Target) {
		//Perform bi-linear interpolation to extract a sub-image from the source image with the given size and from
		//the given location. This function supports types CV_8UC1 and CV_8UC3 only.
	
		//First, if the sample center is too close to an image edge, fall back on standard method
		if ((int(centerX - ((double) X)/2) < 0) || (int(ceil(centerX + ((double) X)/2)) >= Source.rows) ||
		    (int(centerY - ((double) Y)/2) < 0) || (int(ceil(centerY + ((double) Y)/2)) >= Source.cols)) {
		    //fprintf(stderr,"\nResorting to OpenCV Implementation.");
		    getRectSubPix(Source, Size(Y,X), Point2f(centerY, centerX), Target);
		}
		else {
			//Use SIMD to interpolate manually
			int numberOfChannels = Source.channels();
		
			if (numberOfChannels == 3)
				fprintf(stderr,"\n3-channel SIMD is not currently implemented");
			else {
				//1-channel SIMD bilinear interpolation
			
				//Allocate Target buffer, if necessary
				if ((Target.rows != X) || (Target.cols != Y) || (Target.type() != CV_8UC1))
					Target = Mat(X, Y, CV_8UC1);
			
				double xLowStrength  = ceil(centerX) - centerX;
				double yLowStrength  = ceil(centerY) - centerY;
		
				int16_t a =  (int16_t) (yLowStrength  * 128.0);
				int16_t b = ((int16_t) 128) - a;
				int16_t c =  (int16_t) (xLowStrength  * 128.0);
				int16_t d = ((int16_t) 128) - c;
			
				//fprintf(stderr,"\na=%d b=%d c=%d d=%d", a, b, c, d);
			
				__m128i a_Packed = _mm_set1_epi16(a);
				__m128i b_Packed = _mm_set1_epi16(b);
				__m128i c_Packed = _mm_set1_epi16(c);
				__m128i d_Packed = _mm_set1_epi16(d);
			
				//We process 8 elements at a time with current implementation
				__m128i partialBufferMask;
				int numberOfBuffersPerRow = Y/8;
				bool partialBufferNeeded;
				if ((Y % 8) == 0)
					partialBufferNeeded = false;
				else {
					partialBufferNeeded = true;
					numberOfBuffersPerRow++;
					int i=0;
					while (i < (Y % 8))
						((unsigned char *)(&partialBufferMask))[i++] = 255U;
					while (i < 16)
						((unsigned char *)(&partialBufferMask))[i++] = 0U;
				}
			
				__m128i LowerHalfMask;
				for (int i=0; i<8; i++)
					((unsigned char *)(&LowerHalfMask))[i] = 255U;
				for (int i=8; i<16; i++)
					((unsigned char *)(&LowerHalfMask))[i] = 0U;
			
				__m128i shiftAmount = _mm_setzero_si128();
				((unsigned char *)(&shiftAmount))[0] = 7U;
			
				int lowestRow = int(ceil(centerX - ((double) (X - 1))/2)) - 1;
				int lowestCol = int(ceil(centerY - ((double) (Y - 1))/2)) - 1;
			
				//fprintf(stderr,"\nStarting Row = %d, Starting Col = %d", lowestRow, lowestCol);
			
				__m128i Q11, Q12, Q21, Q22, buffer;
				for (int x=0; x<X; x++) {
					char* row1HeadSource = Source.ptr<char>(x + lowestRow)     + lowestCol;
					char* row2HeadSource = Source.ptr<char>(x + lowestRow + 1) + lowestCol;
					char* rowHeadTarget = Target.ptr<char>(x);
					for (int chunk=0; chunk<numberOfBuffersPerRow; chunk++) {
						Q11 = _mm_lddqu_si128((__m128i*)  row1HeadSource);
						Q12 = _mm_lddqu_si128((__m128i*) (row1HeadSource + 1));
						Q21 = _mm_lddqu_si128((__m128i*)  row2HeadSource);
						Q22 = _mm_lddqu_si128((__m128i*) (row2HeadSource + 1));
					
						//Expand the first 8 elements to 16-bit packed integers
						Q11 = _mm_cvtepu8_epi16(Q11);
						Q12 = _mm_cvtepu8_epi16(Q12);
						Q21 = _mm_cvtepu8_epi16(Q21);
						Q22 = _mm_cvtepu8_epi16(Q22);
					
						//Perform Interpolation using integer-scale resolution
						Q11    = _mm_srl_epi16(_mm_adds_epu16(_mm_mullo_epi16(a_Packed, Q11), _mm_mullo_epi16(b_Packed, Q12)), shiftAmount);
						Q21    = _mm_srl_epi16(_mm_adds_epu16(_mm_mullo_epi16(a_Packed, Q21), _mm_mullo_epi16(b_Packed, Q22)), shiftAmount);
						buffer = _mm_srl_epi16(_mm_adds_epu16(_mm_mullo_epi16(c_Packed, Q11), _mm_mullo_epi16(d_Packed, Q21)), shiftAmount);
					
						//Save buffer and increment pointers to proceed to next chunk
						if ((partialBufferNeeded) && (chunk == numberOfBuffersPerRow-1)) {
							 buffer = _mm_packus_epi16(buffer, buffer);                     //Re-pack at 8-bit unsigned ints
							_mm_maskmoveu_si128 (buffer, partialBufferMask, rowHeadTarget); //Save lower 8 elements of buffer to destination
						}
						else {
							 buffer = _mm_packus_epi16(buffer, buffer);                     //Re-pack at 8-bit unsigned ints
							_mm_maskmoveu_si128 (buffer, LowerHalfMask, rowHeadTarget);     //Save lower partial buffer to destination
					
							//Update pointers to point to next chunk
							row1HeadSource += 8;
							row2HeadSource += 8;
							rowHeadTarget  += 8;
						}
					}
				}
			}
		}
	}

	double computeL1Diff_SIMD(Mat& A, Mat& B) {
		//A and B must have the same dims and type. Type must be CV_8UC1 or CV_8UC3
		int X = A.rows;
		int Y = A.cols;
		int numberOfChannels = A.channels();
	
		__m128i partialBufferMask;
		int numberOfFullBuffersPerRow = (Y * numberOfChannels)/16;
		bool partialBufferNeeded;
		if (((Y * numberOfChannels) % 16) == 0)
			partialBufferNeeded = false;
		else {
			partialBufferNeeded = true;
			int i=0;
			while (i < ((Y * numberOfChannels) % 16))
				((unsigned char *)(&partialBufferMask))[i++] = 255U;
			while (i < 16)
				((unsigned char *)(&partialBufferMask))[i++] = 0U;
		}
	
		__m128i accum = _mm_setzero_si128();
		__m128i temp1, temp2;
		for (int x=0; x<X; x++) {
			__m128i* rowHead_a = A.ptr<__m128i>(x);
			__m128i* rowHead_b = B.ptr<__m128i>(x);
		
			for (int chunk=0; chunk<numberOfFullBuffersPerRow; chunk++) {
				accum = _mm_add_epi64(_mm_sad_epu8(_mm_lddqu_si128(rowHead_a + chunk), _mm_lddqu_si128(rowHead_b + chunk)), accum);
			}
			if (partialBufferNeeded) {
				temp1 = _mm_blendv_epi8(partialBufferMask, _mm_lddqu_si128(rowHead_a + numberOfFullBuffersPerRow), partialBufferMask);
				temp2 = _mm_blendv_epi8(partialBufferMask, _mm_lddqu_si128(rowHead_b + numberOfFullBuffersPerRow), partialBufferMask);
				accum = _mm_add_epi64(_mm_sad_epu8(temp1, temp2), accum);
			}
		}
		return((double)(_mm_extract_epi64(accum, 0) + _mm_extract_epi64(accum, 1)));
	}
	
	
	void UIntMatToFloatMat(Mat& In, Mat& Out) {
		//Hardware accelerated conversion from UInt8 type matrix to Float32 type matrix
		
		//This function does not seem to provide any speedup at all over serial conversion
		//This is probably due to a memory bandwidth bottleneck.
		
		int X = In.rows;
		int Y = In.cols;
		int numberOfChannels = In.channels();
		
		//Re-allocate out matrix, if necessary
		if ((Out.rows != X) || (Out.cols != Y) || (Out.depth() != CV_32F) || (Out.channels() != numberOfChannels))
			Out = Mat(X, Y, CV_32FC(numberOfChannels));
		
		//Compute the number of full buffers per row and the number left over in each row
		int numberOfFullBuffersPerRow = (Y * numberOfChannels)/8;
		int numberOfElementsLeftOverPerRow;
		bool partialBufferNeeded;
		
		numberOfElementsLeftOverPerRow = (Y * numberOfChannels) % 8;
		if (numberOfElementsLeftOverPerRow == 0)
			partialBufferNeeded = false;
		else
			partialBufferNeeded = true;
		
		__m128i uint8Buffer;
		__m128i lower_int32, upper_int32;
		__m256i buffer_int32;
		__m256 buffer_float;
		
		for (int x=0; x<X; x++) {
			__m128i* rowHead_In  = In.ptr<__m128i>(x);
			__m256*  rowHead_Out = Out.ptr<__m256>(x);
			
			for (int chunk=0; chunk<numberOfFullBuffersPerRow; chunk++) {
				//Load 8 uints
				uint8Buffer = _mm_lddqu_si128 ((__m128i*)(((char *)rowHead_In) + 8*chunk)); //Only use lower 8 items
		
				//Convert to 8 floats
				lower_int32 = _mm_cvtepu8_epi32 (uint8Buffer);    //4 uint8 -> 4 int32 (convert low 4 elements)
				uint8Buffer = _mm_srli_si128 (uint8Buffer, 4);    //Shift uint8 buffer right by 4 bytes
				upper_int32 = _mm_cvtepu8_epi32 (uint8Buffer);    //4 uint8 -> 4 int32 (convert low 4 elements)
				buffer_int32 = _mm256_castsi128_si256 (lower_int32);
				buffer_int32 = _mm256_insertf128_si256 (buffer_int32, upper_int32, 1);
		
				buffer_float = _mm256_cvtepi32_ps (buffer_int32); //8 int32 -> 8 floats
		
				//save floats
				_mm256_storeu_ps ((float*)(rowHead_Out + chunk), buffer_float);
			}
			if (partialBufferNeeded) {
				uint8_t * partialBufferHead_in = (uint8_t *)(((char *)rowHead_In) + 8*numberOfFullBuffersPerRow);
				float * partialBufferHead_out = (float *)(rowHead_Out + numberOfFullBuffersPerRow);
				for (int itemCounter = 0; itemCounter < numberOfElementsLeftOverPerRow; itemCounter++) {
					partialBufferHead_out[itemCounter] = (float)(partialBufferHead_in[itemCounter]);
				}
			}
		}
	}
	
	int min_int(int x, int y) {
		if (x < y)
			return(x);
		else
			return(y);
	}
	
	double coputeStandardDeviation(vector<double> &errors, vector<uchar> &pointUseFlags) {
		//This function computes the standard deviation of the values in "errors", ignoring all
		//values for which the corresponding pointUseFlag is 0. "errors" is assumed to contain
		//samples from a 0-mean process.
	
		if (errors.size() != pointUseFlags.size()) {
			fprintf(stderr,"coputeStandardDeviation: errors and pointUseFlags don't have the same sizes! Aborting.\n\r");
			fprintf(stderr,"errors has size %d, while pointUseFlags has size %d.\n\r", (int) errors.size(), (int) pointUseFlags.size());
			return(0.0);
		}
	
		double cumsum = 0.0;
		int numberOfPointsUsed = 0;
		for (int i=0; i<((int) pointUseFlags.size()); i++) {
			if (pointUseFlags[i] != 0) {
				cumsum += errors[i]*errors[i];
				numberOfPointsUsed++;
			}
		}
		if (numberOfPointsUsed > 0) return(sqrt(cumsum/numberOfPointsUsed));
		else return(0.0);
	}

	double coputeStandardDeviation(vector<double> &errors) {
		//This function computes the standard deviation of the values in "errors".
		//"errors" is assumed to contain samples from a 0-mean process.
		double cumsum = 0.0;
		for (int i=0; i<((int) errors.size()); i++) {
			cumsum += errors[i]*errors[i];
		}
		if (errors.size() > 0) return(sqrt(cumsum/errors.size()));
		else return(0.0);
	}
	
	void homogeneousLinearLeastSquares(MatrixXd &A, VectorXd &v, VectorXd &sigma) {
		//This function finds the unit vector v that best solves the system Av=0 (in a least-squqres sense).
		//It also returns the singular values of A. Ideally, the smallest singular value
		//will be near 0, and all others will be far from 0. In this case, there is effectively
		//one "good" unit vector v that solves Av=0. If there are multiple singular values that are
		//near 0, then there are multiple orthogonal vectors that do a "good" job solving Av=0. The one
		//that solves it best is still returned, but the algorithm will be somewhat unstable. That is,
		//small changes in the input matrix could result in a significantly different v being returned.
		//If no singular value of A are near 0, then no unit vectors do a good job of solving Av=0,
		//but the one that minimizes the MSE residuals is still returned.
		//
		//The singular values are returned as a vector "sigma" sorted in decreasing order.
		//v and sigma will be allocated if necessary.
	
		int d = A.cols();
		if (A.rows() < d) {
			fprintf(stderr,"homogeneousLinearLeastSquares: A has too few rows to detirmine best v. Aborting.");
			return;
		}
	
		MatrixXd V;
		JacobiSVD<MatrixXd> svd(A, ComputeThinV);
		V = svd.matrixV();
		sigma = svd.singularValues(); //Copy singular values of A into sigma
	
		//Allocate v and sigma if necessary
		if (v.rows() != d) v = VectorXd(d);
		if (sigma.rows() != d) sigma = VectorXd(d);
	
		//Copy right singular vector of A into v.
		for (int i=0; i<d; i++) {
			v(i) = V(i,d-1);
		}
	}
	
	bool isFirstNewer(struct timeval a, struct timeval b) {
		//Returns true if "a" corresponds to a time after "b" (or the same instant).
		//Return false otherwise.
		if (timeDiff(a, b) >= 0.0) return(true);
		else return(false);
	}
	
	double timeDiff(struct timeval a, struct timeval b) {
		//Returns elapsed time (in seconds) from b to a (i.e a-b).
		//If a occurred prior to b, the result will be negative.
		long int sec, usec;
		
		sec = a.tv_sec - b.tv_sec;
		usec = a.tv_usec - b.tv_usec;
		return(((double) sec) + 1.0e-6*((double) usec));
	}
	
	void moveForwardsInTime(struct timeval &a, double seconds) {
		//Alter a time struct in place to advance it forwards in time
		//by "seconds" seconds. "seconds" can be positive or negative.
		if (seconds > 0.0) {
			a.tv_sec  += (int)seconds;
			a.tv_usec += (int)(fmod(seconds, 1.0)*1000000);
			if (a.tv_usec >= 1000000) {
				a.tv_sec++;
				a.tv_usec -= 1000000;
			}
		}
		else if (seconds < 0.0) {
			seconds = -1.0*seconds;
			int fullSeconds  = (int)seconds;
			int microseconds = (int)(fmod(seconds, 1.0)*1000000);
			
			if (a.tv_usec >= microseconds) {
				a.tv_sec  -= fullSeconds;
				a.tv_usec -= microseconds;
			}
			else {
				a.tv_sec  -= (fullSeconds + 1);
				a.tv_usec += (1000000 - microseconds);
			}
		}
	}
	
	void moveBackwardsInTime(struct timeval &a, double seconds) {
		//Alter a time struct in place to move it backwards in time
		//by "seconds" seconds.
		moveForwardsInTime(a, -1.0*seconds);
	}
	
	vector<string> glob(const string& pat) {
		//Taken from random (unsourced) place online
		glob_t glob_result;
		glob(pat.c_str(),GLOB_TILDE,NULL,&glob_result);
		vector<string> ret;
		for(unsigned int i=0;i<glob_result.gl_pathc;++i){
		   ret.push_back(string(glob_result.gl_pathv[i]));
		}
		globfree(&glob_result);
		return ret;
	}
	
	void drawCenteredText(Mat &Image, string& text, Point location, double size, int thickness, int fontFace, Scalar color) {
		int baseline=0;
		Size textSize = getTextSize(text, fontFace, size, thickness, &baseline);
		baseline += thickness;
	
		//Center the text
		Point textOrg(location.x - textSize.width/2, location.y + textSize.height/2);
	
		//Draw the box
		rectangle(Image, textOrg + Point(-thickness, thickness), textOrg + Point(textSize.width + thickness, -textSize.height - 3*thickness), color, thickness);
	
		//Draw Text
		putText(Image, text, textOrg, fontFace, size, Scalar::all(0), thickness, 8);
		putText(Image, text, textOrg, fontFace, size, color, thickness/2, 8);
	}
	
	int countDown(VideoCapture cap, const char* windowName, int count, string line1, string line2) {
		//Render a countdown overlayed with the output from the provided video capture device.
		//The countdown starts at "count" and ticks down very 1/2 second until 0 is reached.
		//"line1" and "line2" are also rendered to the screen, if not empty strings.
		//If the countdown is terminated by pressing spacebar, -1 is returned. Otherwise, 0 is returned.
		Mat frame;
		int key;
		int renderLines = 1;
		if ((line1.empty()) && (line2.empty())) renderLines = 1;
		else if ((!line1.empty()) && (!line2.empty())) renderLines = 3;
		else if (line1.empty()) {
			renderLines = 2;
			line1 = line2;
		}
		else {
			renderLines = 2;
		}
	
		for (int tick = count; tick >= 0; tick--) {
			for (int frameNumber=0; frameNumber<15; frameNumber++) {
				cap >> frame;
				string numberText = to_string(tick);
			
				if (renderLines == 1) {
					drawCenteredText(frame, numberText,   Point(frame.cols/2, frame.rows/2+20), 2, 2, FONT_HERSHEY_PLAIN);
				}
				else if (renderLines == 2) {
					drawCenteredText(frame, line1, Point(frame.cols/2, frame.rows/2-20), 2, 2, FONT_HERSHEY_PLAIN);
					drawCenteredText(frame, numberText,   Point(frame.cols/2, frame.rows/2+20), 2, 2, FONT_HERSHEY_PLAIN);
				}
				else {
					drawCenteredText(frame, line1, Point(frame.cols/2, frame.rows/2-40), 2, 2, FONT_HERSHEY_PLAIN);
					drawCenteredText(frame, line2, Point(frame.cols/2, frame.rows/2), 2, 2, FONT_HERSHEY_PLAIN);
					drawCenteredText(frame, numberText,   Point(frame.cols/2, frame.rows/2+40), 2, 2, FONT_HERSHEY_PLAIN);
				}
			
				imshow(windowName, frame); //Render Countdown
				key = waitKey(1);
				if (key == 32) return(-1);
			}
		}
		return(0);
	}

	int waitForButtonPress(VideoCapture cap, const char* windowName, int count, string line1, string line2) {
		//Render text overtop of screen until a button is pressed. When a button is pressed, the function returns to pressed key.
		Mat frame;
		int key;
		int renderLines = 1;
		if ((line1.empty()) && (line2.empty())) renderLines = 1;
		else if ((!line1.empty()) && (!line2.empty())) renderLines = 3;
		else if (line1.empty()) {
			renderLines = 2;
			line1 = line2;
		}
		else {
			renderLines = 2;
		}
	
		string messageText = "Press any key to continue";
	
		while(1) {
			cap >> frame;
		
			if (renderLines == 1) {
				drawCenteredText(frame, messageText,   Point(frame.cols/2, frame.rows/2+20), 2, 2, FONT_HERSHEY_PLAIN);
			}
			else if (renderLines == 2) {
				drawCenteredText(frame, line1, Point(frame.cols/2, frame.rows/2-20), 2, 2, FONT_HERSHEY_PLAIN);
				drawCenteredText(frame, messageText,   Point(frame.cols/2, frame.rows/2+20), 2, 2, FONT_HERSHEY_PLAIN);
			}
			else {
				drawCenteredText(frame, line1, Point(frame.cols/2, frame.rows/2-40), 2, 2, FONT_HERSHEY_PLAIN);
				drawCenteredText(frame, line2, Point(frame.cols/2, frame.rows/2), 2, 2, FONT_HERSHEY_PLAIN);
				drawCenteredText(frame, messageText,   Point(frame.cols/2, frame.rows/2+40), 2, 2, FONT_HERSHEY_PLAIN);
			}
		
			imshow(windowName, frame); //Render Countdown
			key = waitKey(1);
			if (key != -1) return(key);
		}
	}
	
	
	int saveTrajectories(FeatureSet &features, std::string outputFilePath) {
		//Save the trajectories of a feature set to a text file with path outputFilePath.
		//Upon success, this function returns 0. If the target file cannot be written, the function
		//fails and returns -1. The trajectories are written in the following format. Each feature is
		//written to its own row. A single row has form (f,x,y):(f,x,y): ... :(f,x,y)
		//where f is a frame number, and x and y are the vertical and horizontal positions of the
		//feature in frame f, respectively (this is opposite openCV standard convention). This format
		//allows for representing sets of features that exist at different times in the video sequence.
		//In this format, blank lines and lines beginning with # are to be ignored.
		
		FILE *fid;
		fid = fopen(outputFilePath.c_str(), "w");
		if (fid == NULL) return(-1);
		
		for (list<Feature>::iterator featureIter = features.Features.begin(); featureIter != features.Features.end(); featureIter++) {
			//Write the trajectory of the current feature to the file
			Feature f = *featureIter;
			
			if ((int) f.History.size() > 0) {
				int frameNumber = f.BirthFrameNumber;
				list<cv::Vec2f>::iterator locIter     = f.History.begin();
				list<cv::Vec2f>::iterator nextLocIter = locIter;
				nextLocIter++;
				while (nextLocIter != f.History.end()) {
					Vec2f location = *locIter;
					fprintf(fid, "(%d, %.15f, %.15f):", frameNumber, location[0], location[1]);
					frameNumber++;
					locIter++;
					nextLocIter++;
				}
				Vec2f location = *locIter;
				fprintf(fid, "(%d, %.15f, %.15f)\n", frameNumber, location[0], location[1]);
			}
		}
		fclose(fid);
		return(0);
	}
	
	int loadTrajectories(string inputFilePath, list<int> &birthFrameNumbers, list<list<Vec2f>> &trajectories) {
		//Load trajectories from a file in the format produced by the function "saveTrajectories".
		
		fstream myFile;
		myFile.open(inputFilePath.c_str() , ios::in);
		
		if (!myFile.is_open())
			return(-1);
		
		//Flush the target lists
		birthFrameNumbers.clear();
		trajectories.clear();
		
		string currentLine;
		while(true) {
			if (!getline(myFile, currentLine)) {
				//We could not read more data from the file
				if (myFile.eof())
					//There is no problem, we just reached the end of the file.
					break;
				else {
					myFile.close();
					return(-1);
				}
			}
			
			int firstCharIndex = currentLine.find_first_not_of(' ');
			if ((size_t) firstCharIndex == string::npos)
				//This line is empty - ignore it
				continue;
			
			if (currentLine.at(firstCharIndex) == '#')
				//This is a comment line - ignore it
				continue;
			
			//Start parsing the cells in the line
			int cellStartIndex = 0;
			int cellEndIndex = 0;
			
			list<int> frameNumbers;
			list<Vec2f> locations;
			
			while (true) {
				cellStartIndex = currentLine.find_first_of('(', cellEndIndex);
				if ((size_t) cellStartIndex == string::npos)
					//There are no more cells in this line
					break;
				cellEndIndex = currentLine.find_first_of(')', cellStartIndex);
				if ((size_t) cellEndIndex == string::npos) {
					//There is no terminating bracket for this cell
					myFile.close();
					return(-1);
				}
				
				int firstCommaIndex = currentLine.find_first_of(',', cellStartIndex);
				int secondCommaIndex = currentLine.find_first_of(',', firstCommaIndex+1);
				if (((size_t) firstCommaIndex == string::npos) || ((size_t) secondCommaIndex == string::npos)) {
					//Cell is invalid - it does not contain 2 commas
					myFile.close();
					return(-1);
				}
					
				int frameNum = stoi(currentLine.substr(cellStartIndex+1, firstCommaIndex - cellStartIndex - 1));
				float x = stof(currentLine.substr(firstCommaIndex+1, secondCommaIndex - firstCommaIndex - 1));
				float y = stof(currentLine.substr(secondCommaIndex+1, cellEndIndex - secondCommaIndex - 1));
				
				Vec2f loc(x,y);
				frameNumbers.push_back(frameNum);
				locations.push_back(loc);
			}
			
			if ((int)(frameNumbers.size()) == 0)
				//The line contained no valid cells - move on
				continue;
			
			int birthFrame = frameNumbers.front();
			birthFrameNumbers.push_back(birthFrame);
			trajectories.push_back(locations);
		}
		
		myFile.close();
		return(0);
	}
	
	int loadInitialPositions(std::string inputFilePath, std::list<cv::Vec2f> &initPositions) {
		//Load initial feature positions from an "Initial Positions File". The data is returned as a list of Vec2f objects.
		fstream myFile;
		myFile.open(inputFilePath.c_str() , ios::in);
		
		if (!myFile.is_open())
			return(-1);
		
		//Flush the target list
		initPositions.clear();
		
		string currentLine;
		while(true) {
			if (!getline(myFile, currentLine)) {
				//We could not read more data from the file
				if (myFile.eof())
					//There is no problem, we just reached the end of the file.
					break;
				else {
					myFile.close();
					return(-1);
				}
			}
			
			int firstCharIndex = currentLine.find_first_not_of(' ');
			if ((size_t) firstCharIndex == string::npos)
				//This line is empty - ignore it
				continue;
			
			if (currentLine.at(firstCharIndex) == '#')
				//This is a comment line - ignore it
				continue;
			
			//Start parsing the cells in the line
			int cellStartIndex = currentLine.find_first_of('(');
			int cellEndIndex   = currentLine.find_first_of(')');
			if (((size_t) cellStartIndex == string::npos) || ((size_t) cellEndIndex == string::npos)) {
				//Line is mal-formed
				myFile.close();
				return(-1);
			}
			
			int firstCommaIndex  = currentLine.find_first_of(',', cellStartIndex);
			int secondCommaIndex = currentLine.find_first_of(',', firstCommaIndex+1);
			if (((size_t) firstCommaIndex == string::npos) || ((size_t) secondCommaIndex == string::npos)) {
				//Line is mal-formed
				myFile.close();
				return(-1);
			}
			
			int frameNum = stoi(currentLine.substr(cellStartIndex+1, firstCommaIndex - cellStartIndex - 1));
			if (frameNum != 0)
				//Ignore this line
				continue;
			
			float x = stof(currentLine.substr(firstCommaIndex+1, secondCommaIndex - firstCommaIndex - 1));
			float y = stof(currentLine.substr(secondCommaIndex+1, cellEndIndex - secondCommaIndex - 1));
			
			Vec2f location(x,y);
			initPositions.push_back(location);
		}
		
		myFile.close();
		return(0);
	}
}

