//This class includes functions for saving and reading saved imagery and rotation matrices for post-process tracking
#include <fstream>

#include "rawData.h"
#include "boost/filesystem.hpp"
#include "Utilities.h"

using namespace std;
using namespace cv;
using namespace Eigen;
using namespace boost::filesystem;

rawDataSaver::rawDataSaver() {
	//Constructor
	isInitialized = false;
}

int rawDataSaver::init(string DataFolderPath) {
	//This function should be called before trying to save frames or rotation matrices
	//It tries to create the target folder and sets up the object to write data to
	//that folder. Upon success, 0 is returned. Upon failure, -1 is returned.
	folderPath = DataFolderPath;
	
	//Make sure folderPath has a trailing filesep.
	if ((folderPath.back() != '\\') && (folderPath.back() != '/')) {
		boost::filesystem::path slash("/");
		std::string filesep = slash.make_preferred().native();
		folderPath = folderPath + filesep;
	}
	
	//Now try to create the target folder (if it does not already exist)
	//If it already exists or cannot be reated, return failure
	if (exists(folderPath)) return(-1);
	if (create_directories(folderPath)) {
		isInitialized = true;
		return(0);
	}
	else return(-1);
}


int rawDataSaver::saveFrame(int frameNumber, Mat &frame) {
	//Save a single frame as an image in the data folder. The "init" must have been successfully called
	//prior to calling this function. Upon success, 0 is returned. Upon failure, -1 is returned.
	
	if (!isInitialized) return(-1);
	bool success;
	string filePath;
	filePath = folderPath + string("Frame_") + to_string(frameNumber) + string(".png");
	try {
		success = imwrite(filePath, frame);
	}
	catch (...) {
		success = false;
	}
	if (success) return(0);
	else return(-1);
}

int rawDataSaver::saveRotation(int startFrameNumber, int endFrameNumber, Matrix3d &R) {
	//Save a single rotation matrix to the data folder. The "init" must have been successfully called
	//prior to calling this function. endFrameNumber must be 1 greater than startFrameNumber.
	//Upon success, 0 is returned. Upon failure, -1 is returned.
	
	if (!isInitialized) return(-1);
	string filePath = folderPath + string("Rotation_") + to_string(startFrameNumber) + string("_") + to_string(endFrameNumber) + string(".txt");
	FILE *fid;
	fid = fopen(filePath.c_str(), "w");
	if (fid == NULL) return(-1);
	fprintf(fid, "%.15f, %.15f, %.15f\n", R(0,0), R(0,1), R(0,2));
	fprintf(fid, "%.15f, %.15f, %.15f\n", R(1,0), R(1,1), R(1,2));
	fprintf(fid, "%.15f, %.15f, %.15f",   R(2,0), R(2,1), R(2,2));
	fclose(fid);
	return(0);
}

string rawDataSaver::getNextAvailableFolderPath(string basePath) {
	//This function takes a folder path as an input argument.
	//It looks for folders with the same base path but with a
	//trailing number. It returns the first unused path of the
	//form basepath + " " + n where n is a non-negative integer.
	//This function will only search for through 1000 paths before
	//aborting and returning an empty string. The object does not
	//need to be initialized prior to calling this function.
	//The returned path will not end with a trailing filesep.
	
	//Make sure basePath does not have a trailing filesep.
	if ((basePath.back() == '\\') || (basePath.back() == '/')) {
		basePath = basePath.substr(0, basePath.size()-1);
	}
	
	bool freePathFound = false;
	string candidatePath;
	for (int n=0; n<1000; n++) {
		candidatePath = basePath + string(" ") + to_string(n);
		if (!exists(candidatePath)) {
			freePathFound = true;
			break;
		}
	}
	
	if (freePathFound) return(candidatePath);
	else return(string(""));
}
 
 

rawDataLoader::rawDataLoader() {
	//Constructor
	isInitialized = false;
	numberOfFramesAvailable = 0;
	allRotationsPresent = false;
}

int rawDataLoader::init(std::string DataFolderPath) {
	//This function initializes the raw data loader object. This should be called
	//before calling any other methods of the class. This will return 0 if the provided
	//folder exists and appears to contain valid imagery. It will return -1 otherwise.
	folderPath = DataFolderPath;
	
	//Make sure folderPath has a trailing filesep.
	if ((folderPath.back() != '\\') && (folderPath.back() != '/')) {
		boost::filesystem::path slash("/");
		std::string filesep = slash.make_preferred().native();
		folderPath = folderPath + filesep;
	}
	
	//Now check to see if the directory exists and examine contents
	if (!exists(folderPath)) return(-1);
	
	vector<string> FrameFilenameList = Utilities::glob(folderPath + string("Frame_*.png"));
	numberOfFramesAvailable = (int)(FrameFilenameList.size());
	
	if (numberOfFramesAvailable == 0) {
		
		fprintf(stderr,"Error: Provided folder contains no valid frame images.\n\r");
		return(-1);
	}
	
	list<int> frameNumbersList;
	for (int i=0; i<numberOfFramesAvailable; i++) {
		string fileName = FrameFilenameList[i];
		int ind1 = fileName.find_last_of('_');
		int ind2 = fileName.find_last_of('.');
		int frameNumber = stoi(fileName.substr(ind1+1, ind2-ind1-1));
		frameNumbersList.push_front(frameNumber);
	}
	vector<int> frameNumbers(frameNumbersList.begin(), frameNumbersList.end());
	std::sort(frameNumbers.begin(), frameNumbers.end());
	
	bool allframesPresent = true;
	allRotationsPresent = true;
	for (int i=0; i<(numberOfFramesAvailable-1); i++) {
		if (frameNumbers[i+1] != (frameNumbers[i] + 1))
			allframesPresent = false;
		string filePath = folderPath + string("Rotation_") + to_string(frameNumbers[i]) + string("_") + to_string(frameNumbers[i+1]) + string(".txt");
		if (!exists(filePath))
			allRotationsPresent = false;
	}
	
	if (!allframesPresent) {
		fprintf(stderr,"Error: Frame numbers in provided folder do not form a contiguous sequence of integers.\n\r");
		//For debug support, print out provided frame numbers
		fprintf(stderr,"Frames found:\n\r");
		for (int i=0; i<((int) frameNumbers.size()); i++)
			fprintf(stderr,"Frame %4d\n\r",frameNumbers[i]);
		return(-1);
	}
	
	isInitialized = true;
	return(0);
}

int rawDataLoader::getNumberOfFramesAvailable(void) {
	//Returns the number of frames found in the source data folder, or -1 if the object has not been successfully initialized.
	if (!isInitialized) return(-1);
	else return(numberOfFramesAvailable);
}

bool rawDataLoader::containsRotations(void) {
	//Returns true if the data folder appears to contain rotation data between each pair of consecutive frames in the folder.
	//Returns false if any rotation(s) are missing, or if the object has not been successfully initialized.
	return(allRotationsPresent);
}

int rawDataLoader::getFrame(int frameNumber, cv::Mat &frame) {
	//Loads the desired frame into "frame". This method returns 0 upon success and -1 upon failure.
	if (!isInitialized) return(-1);
	string filePath = folderPath + string("Frame_") + to_string(frameNumber) + string(".png");
	try {
		frame = imread(filePath.c_str());
	}
	catch (...) {
		return(-1);
	}
	return(0);
}

int rawDataLoader::getRotation(int startFrameNumber, int endFrameNumber, Eigen::Matrix3d &R) {
	//Loads the rotation between frame "startFrameNumber" and frame "endFrameNumber" into R.
	//"endFrameNumber" must be exactly 1 greated than "startFrameNumber". Upon success, this method returns 0.
	//It returns -1 on failure, or if the object has not been successfully initialized yet.
	if (!isInitialized) return(-1);
	string filePath = folderPath + string("Rotation_") + to_string(startFrameNumber) + string("_") + to_string(endFrameNumber) + string(".txt");
	
	string numberText;
	fstream myFile;
	myFile.open(filePath.c_str(), ios::in);
	if (!myFile.is_open()) return(-1);
	
	try {
		getline(myFile, numberText, ',');
		R(0,0) = stof(numberText.c_str());
		getline(myFile, numberText, ',');
		R(0,1) = stof(numberText.c_str());
		getline(myFile, numberText);
		R(0,2) = stof(numberText.c_str());
	
		getline(myFile, numberText, ',');
		R(1,0) = stof(numberText.c_str());
		getline(myFile, numberText, ',');
		R(1,1) = stof(numberText.c_str());
		getline(myFile, numberText);
		R(1,2) = stof(numberText.c_str());
	
		getline(myFile, numberText, ',');
		R(2,0) = stof(numberText.c_str());
		getline(myFile, numberText, ',');
		R(2,1) = stof(numberText.c_str());
		getline(myFile, numberText);
		R(2,2) = stof(numberText.c_str());
	}
	catch (...) {
		return(-1);
	}
	myFile.close();
	return(0);
}


